/// <reference types="bluebird" />
import * as Promise from 'bluebird';
import { IAddress, IMessage, Session, UniversalBot } from 'botbuilder';
import { IConfig, MessageFilter } from './config';
import { PossibleExpectedMessageType } from './ExpectedMessage';
export declare type checkSessionFunction = (s: Session) => void;
/**
 * methods that the Test builder can call to edit/modify the options. These can be called until any of the IBotTester method are called.
 */
export interface IOptionsModifier {
    /**
     * adds a message filter to the options.
     * @param filter message filter to add
     */
    addMessageFilter(filter: MessageFilter): BotTester;
    /**
     * sets the timeout time that the BotTester will wait for any particular message before failing
     */
    setTimeout(milliseconds: number): BotTester;
    /**
     * adds prebuilt filter to ignore typing event
     */
    ignoreTypingEvent(): BotTester;
    /**
     * adds prebuilt filter to ignore endOfConversationEvent
     */
    ignoreEndOfConversationEvent(): BotTester;
}
/**
 * Test builder/runner suite. After any of these are called, no functions in IConfigModified should be accessible
 */
export interface IBotTester {
    /**
     * executes each test step serially.
     */
    runTest(): Promise<{}>;
    /**
     * loads a session associated with an address and passes it to a user defined function
     * @param sessionCheckerFunction function passed in to inspect message
     * @param address (Optional) address of the session to load. Defaults to bot's default address if not defined
     */
    checkSession(sessionCheckerFunction: checkSessionFunction, address?: IAddress): IBotTester;
    /**
     * sends a message to a bot and compares bot responses against expectedResponsess. Expected responses can be a variable number of args,
     * each of which can be a single expected response of any PossibleExpectedMessageType or a collection of PossibleExpectedMessageType
     * that mocks a randomly selected response by the bot
     * @param msg message to send to bot
     * @param expectedResponses (Optional) responses the bot-tester framework checks against
     */
    sendMessageToBot(msg: IMessage | string, ...expectedResponses: (PossibleExpectedMessageType | PossibleExpectedMessageType[])[]): IBotTester;
    /**
     * same as sendMessageToBot, but the order of responses is not checked. This will cause the test to hang until all messages it expects
     * are returned
     */
    sendMessageToBotIgnoringResponseOrder(msg: IMessage | string, ...expectedResponses: (PossibleExpectedMessageType | PossibleExpectedMessageType[])[]): IBotTester;
    /**
     * sends a message to the bot. This should be used whenever session.save() is used without sending a reply to the user. This exists due
     * to a limitation in the current implementation of the botbuilder framework
     *
     * @param msg message to send to bot
     */
    sendMessageToBotAndExpectSaveWithNoResponse(msg: IMessage | string): IBotTester;
    /**
     * Works exactly like Promise's .then function, except that the return value is not passed as an arg to the next function (even if its
     * another .then)
     * @param fn some function to run
     */
    then(fn: Function): IBotTester;
    /**
     * Waits for the given delay between test steps.
     * @param delayInMiliseconds time to wait in milliseconds
     */
    wait(delayInMilliseconds: number): IBotTester;
}
/**
 * Test builder and runner for botbuilder bots
 */
export declare class BotTester implements IBotTester, IOptionsModifier {
    private bot;
    private sessionLoader;
    private messageService;
    private testSteps;
    private config;
    constructor(bot: UniversalBot, options?: IConfig);
    addMessageFilter(messageFilter: MessageFilter): BotTester;
    setTimeout(milliseconds: number): BotTester;
    ignoreEndOfConversationEvent(): BotTester;
    ignoreTypingEvent(): BotTester;
    /**
     * Initializes the MessegeService here to allow config changes to accumulate
     */
    runTest(): Promise<{}>;
    checkSession(sessionCheckerFunction: checkSessionFunction, address?: IAddress): IBotTester;
    sendMessageToBot(msg: IMessage | string, ...expectedResponses: (PossibleExpectedMessageType | PossibleExpectedMessageType[])[]): IBotTester;
    sendMessageToBotIgnoringResponseOrder(msg: IMessage | string, ...expectedResponses: (PossibleExpectedMessageType | PossibleExpectedMessageType[])[]): IBotTester;
    sendMessageToBotAndExpectSaveWithNoResponse(msg: IMessage | string): IBotTester;
    then(fn: Function): IBotTester;
    wait(delayInMilliseconds: number): IBotTester;
    private convertToIMessage(msg);
    /**
     * Packages the expected messages into an ExpectedMessage collection to be handed off to the MessageService's sendMessageToBot function
     * @param message message to be sent to bot
     * @param expectedResponses expected responses
     */
    private sendMessageToBotInternal(message, expectedResponses, ignoreOrder?);
}
