"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const Promise = require("bluebird");
const botbuilder_1 = require("botbuilder");
const builtInMessageFilters_1 = require("./builtInMessageFilters");
const config_1 = require("./config");
const ExpectedMessage_1 = require("./ExpectedMessage");
const MessageService_1 = require("./MessageService");
const SessionService_1 = require("./SessionService");
/**
 * Test builder and runner for botbuilder bots
 */
class BotTester {
    constructor(bot, options = config_1.config) {
        this.config = Object.assign({}, config_1.config, options);
        this.config.messageFilters = this.config.messageFilters.slice();
        this.bot = bot;
        this.messageService = new MessageService_1.MessageService(bot, this.config);
        this.sessionLoader = new SessionService_1.SessionService(bot);
        this.testSteps = [];
    }
    addMessageFilter(messageFilter) {
        this.config.messageFilters.push(messageFilter);
        return this;
    }
    setTimeout(milliseconds) {
        this.config.timeout = milliseconds;
        return this;
    }
    ignoreEndOfConversationEvent() {
        this.config.ignoreEndOfConversationEvent = true;
        return this;
    }
    ignoreTypingEvent() {
        this.config.ignoreTypingEvent = true;
        return this;
    }
    /**
     * Initializes the MessegeService here to allow config changes to accumulate
     */
    runTest() {
        if (this.config.ignoreTypingEvent) {
            this.config.messageFilters.push(builtInMessageFilters_1.ignoreTypingEventFilter);
        }
        if (this.config.ignoreEndOfConversationEvent) {
            this.config.messageFilters.push(builtInMessageFilters_1.ignoreEndOfConversationEventFilter);
        }
        this.messageService = new MessageService_1.MessageService(this.bot, this.config);
        return Promise.mapSeries(this.testSteps, (fn) => fn());
    }
    checkSession(sessionCheckerFunction, address) {
        const runSessionChecker = () => this.sessionLoader.getSession(address || this.config.defaultAddress)
            .then(sessionCheckerFunction);
        this.testSteps.push(runSessionChecker);
        return this;
    }
    sendMessageToBot(msg, 
        // currently only supports string RegExp IMessage
        ...expectedResponses) {
        const message = this.convertToIMessage(msg);
        // possible that expected responses may be undefined. Remove them
        expectedResponses = expectedResponses.filter((expectedResponse) => expectedResponse);
        return this.sendMessageToBotInternal(message, expectedResponses);
    }
    sendMessageToBotIgnoringResponseOrder(msg, 
        // currently only supports string RegExp IMessage
        ...expectedResponses) {
        const message = this.convertToIMessage(msg);
        // possible that expected responses may be undefined. Remove them
        expectedResponses = expectedResponses.filter((expectedResponse) => expectedResponse);
        return this.sendMessageToBotInternal(message, expectedResponses, true);
    }
    sendMessageToBotAndExpectSaveWithNoResponse(msg) {
        const message = this.convertToIMessage(msg);
        return this.sendMessageToBotInternal(message, [this.sessionLoader.getInternalSaveMessage(message.address)]);
    }
    then(fn) {
        this.testSteps.push(() => Promise.method(fn)());
        return this;
    }
    wait(delayInMilliseconds) {
        this.testSteps.push(() => Promise.delay(delayInMilliseconds));
        return this;
    }
    convertToIMessage(msg) {
        if (typeof (msg) === 'string') {
            return new botbuilder_1.Message()
                .text(msg)
                .address(this.config.defaultAddress)
                .toMessage();
        }
        return msg;
    }
    /**
     * Packages the expected messages into an ExpectedMessage collection to be handed off to the MessageService's sendMessageToBot function
     * @param message message to be sent to bot
     * @param expectedResponses expected responses
     */
    sendMessageToBotInternal(message, 
        // currently only supports string RegExp IMessage
        expectedResponses, ignoreOrder = false) {
        let expectedMessages = [];
        if (!expectedResponses) {
            expectedMessages = [];
        }
        else if (!(expectedResponses instanceof Array)) {
            expectedMessages = [new ExpectedMessage_1.ExpectedMessage(expectedResponses)];
        }
        else if (expectedResponses instanceof Array) {
            if (expectedResponses.length > 0) {
                expectedMessages = expectedResponses
                    .map((currentExpectedResponseCollection) => new ExpectedMessage_1.ExpectedMessage(currentExpectedResponseCollection));
            }
        }
        this.testSteps.push(() => this.messageService.sendMessageToBot(message, expectedMessages, ignoreOrder));
        return this;
    }
}
exports.BotTester = BotTester;
//# sourceMappingURL=BotTester.js.map