import { IMessage, IEvent } from 'botbuilder';
export declare enum ExpectedMessageType {
    String = 0,
    IMessage = 1,
    Regex = 2,
}
/**
 * Types accepted for responses checkers
 */
export declare type PossibleExpectedMessageType = string | IMessage | RegExp | IEvent;
/**
 * Response expectations area always collections. The collection is the set of possible responses, chosen at random. If the collection size
 * is 1, then there is only one response that is expected
 */
export declare type PossibleExpectedMessageCollections = PossibleExpectedMessageType[];
/**
 * Class that wraps expectedResponseCollections for comparison against an outgoing message. One instance of ExpectedMessage represents one
 * response from the bot.
 */
export declare class ExpectedMessage {
    /**
     * set of possible responses, chosen at random. If the collection size is 1, then there is only one response that is expected
     */
    private readonly expectedResponseCollection;
    constructor(expectedResponseCollection: PossibleExpectedMessageType | PossibleExpectedMessageCollections);
    /**
     * routes the outgoingMessage to the proper comparison method based on the expectedResponseCollection. This switch based method exists
     * due to typescript's lack of polymorphism support.
     *
     * @param outgoingMessage outgoing message that is being compared
     */
    checkForMessageMatch(outgoingMessage: IMessage): void;
    toString(): string;
    /**
     * Asserts that outgoingMessage.text is within the expectedResponseStrings collection
     *
     * @param outgoingMessage outgoing message being compared
     * @param expectedResponseStrings collection of possible string values for comparison
     */
    private checkMessageTextForExactStringMatch(outgoingMessage, expectedResponseStrings);
    /**
     * Assumes the expectedResponseCollection are regexs. Asserts that outgoingMessage.text matches at least of the regexs in
     * expectedResponseCollection
     *
     * @param outgoingMessage outgoing message being compared
     */
    private checkMessageTextForRegex(outgoingMessage);
    /**
     * Assumes the expectedResponseCollection is an IMessage[]. Asserts that at least one IMessage in expectedResponseCollection is fully
     * contained within the outgoingMessage (i.e. the outgoingMessage  {type: 'message', text:'hello'}) would be matched successfully
     * against {type: 'message', text: 'hello', user: { id: user1 }} because the outgoing message is contained within the response. It would
     * not successfulyl be matched against {text: 'hello', user: { id: user1 }} because the expected response is missing { type: 'message' }
     * @param outgoingMessage outgoing message being compared
     */
    private deepMessageMatchCheck(outgoingMessage);
}
